GRAB_SPIDER_MODULES = ['spider']

GRAB_PROXY_LIST = {
    'source': '/web/proxy.txt',
    'source_type': 'text_file',
}

# GRAB_CACHE = {
#    'backend': 'mysql',
#    'database': '{{ PROJECT_NAME }}_cache',
#    'user': '',
#    'passwd': '',
# }

# GRAB_THREAD_NUMBER = 15

# GRAB_QUEUE = {
#    'backend': 'redis',
# }

# GRAB_TASK_REFRESH_CACHE = {
#    'foo': True,
# }
